#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>


/* defines importantes para o programa */

#define DIST_HUGU_LUIZ 2      /* Distancia maxima entre Luizinho e Huguinho */
#define TAMANHO_PLANTACAO 10  /* Tamanho da plantacao (quantos sementes serao plantadas)*/


union semun {
   int val;              /* used for SETVAL only */
   struct semid_ds *buf; /* for IPC_STAT and IPC_SET */
   ushort *array;        /* used for GETALL and SETALL */
};


/* funcoes das threads */
void roda_huguinho(int shmid, int semid, int distancia);
void roda_luizinho(int shmid, int semid);
void roda_zezinho(int shmid, int semid);

/* outras funcoes do programa */

/* semaforo */
int inicializaSemaforo();
void fechaSemaforo(int semid);

/* memoria */
int inicializaMemoriaCompartilhada();
void fechaMemoriaCompartilhada(int shmid);


int main(int argc, char**argv) {
   int pid, shmid, semid, status;


   /* criando o ambiente */
   semid = inicializaSemaforo();
   shmid = inicializaMemoriaCompartilhada();

   /* cria Huguinho */
   if( (pid= fork()) < 0) {
      printf("Erro no fork do Huguinho\n");
      exit(1);
   }

   if(pid == 0) {
      roda_huguinho(shmid, semid, DIST_HUGU_LUIZ);
      exit(0);
   }

   /* cria Zezinho */
   if( (pid = fork()) < 0) {
      printf("Erro no fork do Zezinho\n");
      exit(1);
   }
   if(pid == 0) {
      roda_zezinho(shmid, semid);
      exit(0);
   }

   /* cria Luizinhoa */
   if( (pid = fork()) < 0) {
      printf("Erro no fork do Luizinho\n");
      exit(1);
   }
   if(pid == 0) {
      roda_luizinho(shmid, semid);
      exit(0);
   }

   /* espera Luizinho, que eh o ultimo a terminar sua tarefa */
   waitpid(pid, &status, 0);

   fechaSemaforo(semid);
   fechaMemoriaCompartilhada(shmid);

   return 0;

}


   /* inicializa o semaforo e retorna o ID dele */
int inicializaSemaforo() {
   key_t key;
   int semid;
   union semun arg;

   if((key == ftok("plantacao.c", 'J')) == -1) {
      perror("ftok");
      exit(1);
   }

   if((semid = semget(key,1, 0666 | IPC_CREAT)) == -1) {
      perror("semget");
      exit(1);
   }

   arg.val = 1;
   if(semctl(semid, 0, SETVAL, arg) == -1) {
      perror("semctl");
      exit(1);
   }

   return semid;

}

void fechaSemaforo(int semid) {
   union semun arg;

   if(semctl(semid, 0, IPC_RMID, arg) == -1) {
      perror("semctl");
      exit(1);
   }   
}

/* inicializa a memoria compartilhada e rotorna o ID dela */
#define SHM_SIZE 20
int inicializaMemoriaCompartilhada() {
   key_t key;
   int shmid;
   int *data;
   int mode;

   if((key = ftok("luizinho.c", 'R')) == -1) {
      perror("ftok");
      exit(1);
   }

   if((shmid = shmget(key, SHM_SIZE, 0644 | IPC_CREAT)) == -1) {
      perror("shmget");
      exit(1);
   }

   /* inicializando os valores */
   data = shmat(shmid, (void *)0, 0);
   if (data == (int *)(-1)) {
      perror("shmat");
      exit(1);
   }

   /* 0 => posicao do Huguinho; 1 => posicao do Zezinho; 2 => posicao do Luizinho */
   data[0] = data[1] = data[2] = 0;
   /* 3 => tamanho maximo da plantacao */
   data[3] = TAMANHO_PLANTACAO;

   if(shmdt(data) == -1) {
      perror("shmdt");
      exit(1);
   }

   return shmid;
}

void fechaMemoriaCompartilhada(int shmid) {
   struct shmid_ds buf;
   
   if( shmctl(shmid, IPC_RMID,&buf) == -1) {
      perror("shmctl");
      exit(1);
   }

}




void roda_huguinho(int shmid, int semid, int distancia) {
   struct sembuf sb = {0,-1,0};
   int *data;

   data = shmat(shmid, (void *)0, 0);
   if(data == (int *)(-1)) {
      perror("shmat luizinho");
      exit(1);
   }

   /* data[3] ninguem modifica e data[0] apenas Huguinho modifica
      portanto posso ler sem problema algum */
   while(data[0] < data[3]) {
      sb.sem_op = -1;
      /* entrada */
      if(semop(semid, &sb, 1) == 1) {
	 perror("semop luizinho");
	 exit(1);
      }

      /* vez do Huguinho trabalhar -- AREA CRITICA */
      if( data[0] < data[2] + distancia ) {
	 /* huguinho nao pode ficar muito longe do luizinho */
	 data[0]++;
	 printf("[Huguinho] criou o buraco %d\n", data[0]);
      } else {
	 printf("[Huguinho] esta esperando Luizinho tampar buracos\n");
      }
      fflush(NULL);

      /* saida */ 
      sb.sem_op = 1;
      if(semop(semid, &sb, 1) == -1) {
	 perror("semop do luizinho");
	 exit(1);
      }

   }
   printf("Huguinho terminou seu trabalho ...\n");

}

void roda_luizinho(int shmid, int semid) {
   struct sembuf sb = {0,-1,0};
   int *data;

   data = shmat(shmid, (void *)0, 0);
   if(data == (int *)(-1)) {
      perror("shmat luizinho");
      exit(1);
   }

   /* data[3] ninguem modifica, e data[2] apenas luizinho modifica
      portanto posso ler sem problema algum */
   while( data[2] < data[3] ) {

      
      sb.sem_op = -1;
      /* entrada */
      if(semop(semid, &sb, 1) == 1) {
	 perror("semop luizinho");
	 exit(1);
      }
      
      /* vez do luizinho trabalhar -- AREA CRITICA*/
      if(data[2] != data[1]) {
	 /* luizinho tem que tapar um buraco que zezinho jah colocou semente */
	 data[2]++;
	 printf("[Luizinho] tampou o buracao %d\n", data[2]);
      } else {
	 printf("[Luizinho] esta espernado Zezinho colocar semente\n");
      }
      fflush(NULL);

      /* saida */ 
      sb.sem_op = 1;
      if(semop(semid, &sb, 1) == -1) {
	 perror("semop do luizinho");
	 exit(1);
      }


   }
   printf("Luizinho terminou todo o seu trabalho\n");

}


void roda_zezinho(int shmid, int semid) {
   struct sembuf sb = {0,-1,0};
   int *data;

   data = shmat(shmid, (void *)0, 0);
   if(data == (int *)(-1)) {
      perror("shmat zezinho");
      exit(1);
   }

   /* data[3] ninguem modifica, e data[1] apenas zezinho modifica
      portanteo posso ler sem problema nenhum */
   while( data[1] < data[3] ) {
      sb.sem_op = -1;
      /* entrada */
      if(semop(semid, &sb, 1) == 1) {
	 perror("semop luizinho");
	 exit(1);
      }

      /* vez do zezinho trabalhar -- AREA CRITICA */
      if(data[1] != data[0]) {
	 /* zezinho precisa de buraco para colocar a semente */
	 data[1]++;
	 printf("[Zezinho] colocou semente no buraco %d\n", data[1]);
      } else {
	 printf("[Zezinho] esta esperando Huguinho fazer um buraco\n");
      }
      fflush(NULL);

      /* saida */
      sb.sem_op = 1;
      if(semop(semid, &sb, 1) == -1) {
	 perror("semop do luizinho");
	 exit(1);
      }
   }
   printf("Zezinho terminou todo o seu trabalho ...\n");

}

